/**
 * Copyright 2023 Velocity Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const { isNil, omitBy } = require('lodash/fp');
const undici = require('undici');
const CacheableLookup = require('cacheable-lookup');
const { nanoid } = require('nanoid/non-secure');
const pkg = require('../package.json');

const buildContext = (context) => {
  const { log, traceId } = context;

  return omitBy(isNil)({
    log,
    traceId,
    reqId: nanoid(10),
  });
};

const parseRawBody = async (rawBodyBuffer) => {
  const rawBody = await rawBodyBuffer.text();
  try {
    return JSON.parse(rawBody);
  } catch {
    return rawBody;
  }
};

const logResponse = async (
  log,
  responseBody,
  headers,
  statusCode,
  traceId,
  reqId,
  url
) =>
  log.info(
    {
      url,
      reqId,
      traceId,
      responseStatus: statusCode,
      responseHeaders: headers,
      responseBody,
    },
    `Fetch Response: ${statusCode}`
  );

const initClient = ({
  rejectUnauthorized,
  isProd,
  requestTimeout,
  traceIdHeader = 'TRACE_ID',
  customHeaders = {},
  prefixUrl,
}) => {
  const cacheable = new CacheableLookup();
  const clientConfig = {
    connect: {
      rejectUnauthorized: rejectUnauthorized || isProd || false,
      lookup: cacheable.lookup,
    },
    bodyTimeout: requestTimeout,
  };

  const headers = {
    'user-agent': `${pkg.name}/${pkg.version}`,
    ...customHeaders,
  };

  const agent = new undici.Agent(clientConfig);
  undici.setGlobalDispatcher(agent);

  const client = (context) => ({
    get: async (url) => {
      const { log, traceId, reqId } = buildContext(context);

      const httpResponse = await undici.request(`${prefixUrl}${url}`, {
        method: 'GET',
        headers: {
          ...headers,
          [traceIdHeader]: traceId,
        },
        throwOnError: true,
      });

      const response = {
        statusCode: httpResponse.statusCode,
        headers: httpResponse.headers,
        body: await parseRawBody(httpResponse.body),
      };

      await logResponse(
        log,
        response.body,
        response.headers,
        response.statusCode,
        traceId,
        reqId,
        `${prefixUrl}${url}`
      );

      return response;
    },
  });

  return client;
};

module.exports = initClient;
